module HibanaFunctions where

    import HibanaTypes

    getRT :: State -> TMap
    getRT (State (x, _, _, _)) = x

    getV :: State -> VMap
    getV (State (_, x, _, _)) = x

    getP :: State -> PMap
    getP (State (_, _, x, _)) = x

    getF :: State -> FMap
    getF (State (_, _, _, x)) = x

    isNothing :: Maybe a -> Bool
    isNothing Nothing = True
    isNothing _ = False

    isTrue :: T -> Bool
    isTrue (TBool x) = x

    listToNEList :: List a -> NEList a
    listToNEList [x] = One x
    listToNEList (x:xs) = (Cons x (listToNEList xs))

    neListToList :: NEList a -> List a
    neListToList (One a) = [a]
    neListToList (Cons h t) = h:(neListToList t)

    isTError :: T -> Bool
    isTError (TError _) = True
    isTError _ = False

    valid_list :: List T -> Bool
    valid_list [] = True
    valid_list (x:xs) = not (isTError x) && valid_list xs

    getVarType :: State -> ID_v -> ID_t
    getVarType (State (_, (vars, V fun_v), _, _)) v = let Just (t, _) = fun_v v in t

    --Needs checking of array ranges

    isRInt (Just (Right (_, RInt _ _))) = True
    isRInt _ = False
   
    variable_type :: T -> State -> ID_v -> Bool
    variable_type (TError _) _ _ = False
    variable_type (TInt i) env v = let t = getVarType env v
				       (_, RT fun_rt) = getRT env
				       rt = fun_rt t
				   in if isNothing rt then False
					else if isRInt rt then
						let Just (Right (_, RInt i1 i2)) = rt in
						    i >= i1 && i <= i2
					     else False
    variable_type (TBool _) env v = let (ID_t t) = getVarType env v in
				        t=="Boolean"
    variable_type (TArray (ra, _, fun_a)) env v = let t = getVarType env v
						      (_, RT fun_rt) = getRT env
						      rt = fun_rt t
						    in if isNothing rt then False
							else let Just (Right (_, rb)) = rt in rangeArrayEqual ra rb

    rangeArrayEqual :: R -> R -> Bool
    rangeArrayEqual (RArray r11 r12) (RArray r21 r22) = (r11==r21) && (r12==r22)
    rangeArrayEqual _ _ = False
    --This is too strict

    union :: (Eq a) => List a -> List a -> List a
    union [] l2 = l2
    union (x:xs) l2 = if elem x l2 then union xs l2 else union xs (x:l2)
				       
    upd_v :: State -> ID_v -> T -> State
    upd_v (State (t,(lv, V fun_v),p,f)) x k = State (t, (lv, V (\n -> if n==x then (Just (idt, k)) else fun_v n)), p, f)
	where
	Just (idt, val) = fun_v x

    upd_a :: State -> ID_v -> List T -> T -> State
    upd_a (State (t,(lv, V fun_v),p,f)) x le k = State (t, (lv, V (\n -> if n==x then (Just (idt, TArray (r, newlist, newmap))) else fun_v n)), p, f)
		where
		Just (idt, TArray (r, lst, amap)) = fun_v x
		newmap n = if n == (listToNEList le) then k else amap n
		newlist = union [listToNEList le] lst


    addVar :: State -> ID_v -> ID_t -> State
    addVar (State (t,(lv, V fun_v),p,f)) x idt = State (t,(union [x] lv, V (\n -> if n==x then Just (idt, Empty) else fun_v n)),p,f)

    addVarAssign :: State -> ID_v -> ID_t -> T -> State
    addVarAssign (State (t,(lv, V fun_v),p,f)) x idt k = State (t,(union [x] lv, V (\n -> if n==x then Just (idt, k) else fun_v n)),p,f)

    isError :: State -> Bool
    isError (StError _) = True
    isError _ = False
    
    idftoidv :: ID_f -> ID_v
    idftoidv (ID_f f) = ID_v f

    upd_f :: ID_f -> ContractSpec -> ID_t -> Statements -> E -> Parameters_f -> State -> State -> State
    upd_f f c t s e lp envf@(State (af, vf, pf, ff)) envo@(State (ao, vo, po, fo@(funo, F fun_fo))) 
		= State (ao, vo, po, (funo, F new_fun_f))
		where
		new_fun_f x = if x == f then Just (c, lp, t, (vf, ff), s, e) else fun_fo x
    
    addFunctionAndVar :: State -> ID_f -> ID_t -> State
    addFunctionAndVar env@(State (a, (vars, fun_v), p, (funs, fun_f))) f t = State (a, (union [idftoidv f] vars, fun_v), p, (union [f] funs, fun_f))


    addRangeType :: State -> ID_t -> R -> State
    addRangeType env@(State ((lst, RT fun_t), v, p, f)) t r = State ((union [t] lst, RT new_fun_t), v, p, f)
			where
			new_fun_t x = if x==t then Just (Right (Nothing, r)) else fun_t x

    addRangeSubtype :: State -> ID_t -> ID_t -> R -> State
    addRangeSubtype env@(State ((lst, RT fun_t), v, p, f)) t bt r = State ((union [t] lst, RT new_fun_t), v, p, f)
			where
			new_fun_t x = if x==t then Just (Right (Just bt, r)) else fun_t x


    evalExpr :: State -> E -> T
    evalExpr env E = TError Error
    evalExpr env (Atom t) = t   
    evalExpr env (Var x l) = getVal env x l
    evalExpr env (Boolean be) = evalBExpr env be
    evalExpr env (K k) = evalCAExpr k
    evalExpr env (Sum e1 e2) = operatorsT TSum (evalExpr env e1) (evalExpr env e2)
    evalExpr env (Mul e1 e2) = operatorsT TMul (evalExpr env e1) (evalExpr env e2)
    evalExpr env (Div e1 e2) = operatorsT TDiv (evalExpr env e1) (evalExpr env e2)
    evalExpr env (Sub e1 e2) = operatorsT TSub (evalExpr env e1) (evalExpr env e2)
    evalExpr env (Paren e) = evalExpr env e
    -- Function Call is missing

    evalBExpr :: State -> BE -> T
    evalBExpr env (BEAtom b) = (TBool b)
    evalBExpr env (BENot be) = let (TBool b) = evalBExpr env be in TBool (not b)
    evalBExpr env (BEParen be) = evalBExpr env be
    evalBExpr env (And b1 b2) = operatorsT TAnd (evalBExpr env b1) (evalBExpr env b2)
    evalBExpr env (Or b1 b2) = operatorsT TOr (evalBExpr env b1) (evalBExpr env b2)
    evalBExpr env (BGT e1 e2) = operatorsT TBGT (evalExpr env e1) (evalExpr env e2)
    evalBExpr env (GTE e1 e2) = operatorsT TGTE (evalExpr env e1) (evalExpr env e2)
    evalBExpr env (BLT e1 e2) = operatorsT TBLT (evalExpr env e1) (evalExpr env e2)
    evalBExpr env (LTE e1 e2) = operatorsT TLTE (evalExpr env e1) (evalExpr env e2)
    evalBExpr env (Eq e1 e2) = operatorsT TEq (evalExpr env e1) (evalExpr env e2)

    evalCAExpr :: CAE -> T
    evalCAExpr CAE = TError Error
    evalCAExpr (KSum k1 k2) = operatorsT TSum (evalCAExpr k1) (evalCAExpr k2)
    evalCAExpr (KMul k1 k2) = operatorsT TMul (evalCAExpr k1) (evalCAExpr k2)
    evalCAExpr (KDiv k1 k2) = operatorsT TDiv (evalCAExpr k1) (evalCAExpr k2)
    evalCAExpr (KSub k1 k2) = operatorsT TSub (evalCAExpr k1) (evalCAExpr k2)
    evalCAExpr (KParen k) = evalCAExpr k
    evalCAExpr (KAtom t) = t

    data TOp = TSum | 
	       TMul | 
	       TDiv | 
	       TSub | 
	       TAnd | 
	       TOr | 
	       TBGT | 
	       TGTE | 
	       TBLT | 
	       TLTE | 
	       TEq
	deriving Show

    operatorsT :: TOp -> T -> T -> T
    operatorsT TSum (TInt i1) (TInt i2) = TInt (i1 + i2)
    operatorsT TSum (TReal f1) (TReal f2) = TReal (f1 + f2)
    operatorsT TMul (TInt i1) (TInt i2) = TInt (i1 * i2)
    operatorsT TMul (TReal f1) (TReal f2) = TReal (f1 * f2)
    operatorsT TDiv (TInt i1) (TInt i2) = TInt (i1 `div` i2)
    operatorsT TDiv (TReal f1) (TReal f2) = TReal (f1 / f2)
    operatorsT TSub (TInt i1) (TInt i2) = TInt (i1 - i2)
    operatorsT TSub (TReal f1) (TReal f2) = TReal (f1 - f2)
    operatorsT TAnd (TBool b1) (TBool b2) = TBool (b1 && b2)
    operatorsT TOr (TBool b1) (TBool b2) = TBool (b1 || b2)
    operatorsT TBGT (TInt i1) (TInt i2) = TBool (i1 > i2)
    operatorsT TBGT (TReal f1) (TReal f2) = TBool (f1 > f2)
    operatorsT TGTE (TInt i1) (TInt i2) = TBool (i1 >= i2)
    operatorsT TGTE (TReal f1) (TReal f2) = TBool (f1 >= f2)
    operatorsT TBLT (TInt i1) (TInt i2) = TBool (i1 < i2)
    operatorsT TBLT (TReal f1) (TReal f2) = TBool (f1 < f2)
    operatorsT TLTE (TInt i1) (TInt i2) = TBool (i1 <= i2)
    operatorsT TLTE (TReal f1) (TReal f2) = TBool (f1 <= f2)
    operatorsT TEq (TInt i1) (TInt i2) = TBool (i1 == i2)
    operatorsT TEq (TReal f1) (TReal f2) = TBool (f1 == f2)
    operatorsT TEq (TBool b1) (TBool b2) = TBool (b1 == b2)
    operatorsT _ _ _ = TError Error

    getVal :: State -> ID_v -> List E -> T
    getVal (State (_, (vars, V fun_v), _, _)) x [] = let res = fun_v x in
							if isNothing res then 
							    TError Error 
							else let Just (_, val) = res in
							    	val		    
    getVal env@(State (a, (vars, V fun_v), b, c)) x le = let res = fun_v x in
							if isNothing res then 
							    TError Error 
							else let Just (_, TArray (_, _, val)) = res 
								 lval = map (evalExpr env) le
							     in if valid_list lval then
								   val (listToNEList lval)
								else TError Error

    evalIndexes :: State -> [E] -> [T]
    evalIndexes env le = map (evalExpr env) le

    reduceState :: State -> State
    reduceState (State (t, (lstv, _), (lstp, _), f)) = State (t, empty_vmap, empty_pmap, empty_fmap)

    addProcedure :: ID_p -> State -> State
    addProcedure idp env@(State (t, v, (lst, fun_p), f)) = State (t, v, (union [idp] lst, fun_p), f)

-- TODO

    type_array_ind :: State -> ID_v -> [T] -> Bool
    type_array_ind env a lk = True

    type_array_val :: State -> ID_v -> T -> Bool
    type_array_val env a k = True

    getEnvF :: State -> ID_f -> State
    getEnvF env f = env

    getEnvP :: State -> ID_p -> State
    getEnvP env p = env

    getBodyFunction :: State -> ID_f -> Statements
    getBodyFunction env f = Null
    
    getBodyProcedure :: State -> ID_p -> Statements
    getBodyProcedure env p = Null

    getReturnExpression :: State -> ID_f -> E
    getReturnExpression env f = E

    restoreLocalVariablesP :: State -> State -> ID_p -> State
    restoreLocalVariablesP env env' p = env

    restoreLocalVariablesF :: State -> State -> ID_f -> State
    restoreLocalVariablesF env env' f = env

    upd_p :: ID_p -> ContractSpec -> Statements -> [Parameters_p] -> State -> State
    upd_p p c s lp env@(State (t, v, (lst, P fun_p), f)) = env

